#!/usr/bin/python2.4

"""
A program to automatically geotag photos
Copyright (C) 2006 Niko Kotilainen

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

# Difference between GPS-time and real time.
# If your camera's clock is for example 1 hour ahead of your gps's clock,
# put 3600 (seconds) here.
timeDifference = 10808

# Change this to your own google maps key.
googleMapsKey  = 'ABQIAAAAosaomZsjWkFRInX49lkSHhTd_rlRcgs51FxcUb_pLrfJQe_FHBQV4wLP3MuD6E0GFGCy5KjzBHlwPA'

# Delete old files in the webpages-directory?
delete=True

# URL for shared files, like custom marker images and
# the video player software.
surl = "http://niko.ws/extras/"

# the target directory is required only if you include video clips
target = "http://niko.ws/geotag/example/"



import Image, math, os, commands
from time import *
from sys import argv

try:
  import psyco
  psyco.full()
except:
  pass

# from http://www.codecomments.com/archive278-2005-8-581860.html
def getThumbnail( filename, size = (100,100) ):
  '''Get a thumbnail image of filename'''
  image = Image.open(filename)
  rx, ry = image.size[0]/float(size[0]), image.size[1]/float(size[1])
  if rx > ry:
    resize = int(size[0]), int(round(image.size[1]*(1.0/rx), 0))
  else:
    resize = int(round(image.size[0]*(1.0/ry), 0)), int(size[1])
  image = image.resize( resize, Image.ANTIALIAS )
  newimage = Image.new( 'RGB', resize, )
  x,y = image.size
  newimage.paste(image, (0,0))#((size[0]-x)/2, (size[1]-y)/2),)
  return newimage


def distance(fr, to):
  if type(fr) == type([]):
    fr, to = getGroupLoc(fr), getGroupLoc(to)

  radlat1 = math.pi * fr["lat"]/180
  radlat2 = math.pi * to["lat"]/180
  radlon1 = math.pi * fr["lon"]/180
  radlon2 = math.pi * to["lon"]/180
  theta = fr["lon"]-to["lon"]
  radtheta = math.pi * theta/180
  #print radlat1, radlat2, radlon1, radlon2, theta, radtheta
  dist = math.sin(radlat1) * math.sin(radlat2) + math.cos(radlat1) * math.cos(radlat2) * math.cos(radtheta)
  #print dist
  if (dist >= 1):
    return 0
  dist = math.acos(dist)
  dist = dist * 180/math.pi
  dist = dist * 60 * 1.852
  return dist


def generateImages():
  jpgFiles = filter((lambda x: x.lower().endswith(".jpg")), os.listdir("."))
  photoDates=[]
  isMovie=[]
  for f in jpgFiles:
    time = commands.getoutput("jhead "+f+ " |grep Date/Time")
    t=mktime(strptime(time, "Date/Time    : %Y:%m:%d %H:%M:%S"))

    if not startTime < t < endTime:
      print f, "not", 
      jpgFiles[jpgFiles.index(f)] = None
      continue
    print "Processing", f, "..."
    #print time
    photoDates.append(t)
    
    movie = f.split('.')[0]
    if (os.path.isfile(movie+'.mov')):
      isMovie.append(True)
      if not os.path.exists("webpages/"+directory+movie+".flv"):
        print "Encoding "+movie+".mov to .flv..."
        encode = "ffmpeg -i "+movie+".mov -s 296x222 -ar 44100 -b 400 webpages/" +directory+movie + ".flv 2>/dev/null"
        addMetaData = "flvtool2 -U webpages/"+directory+movie + ".flv"
        os.system(encode)
        os.system(addMetaData)
      tn=getThumbnail(f, (296,222) )
      tn.save("webpages/"+directory+"small_"+f)
      tn.save("webpages/"+directory+f)
      
    else:
      isMovie.append(False)
    
    for n in [(300,"webpages/"+directory+"small_") , (800,"webpages/"+directory)]:
      newname=n[1]+f
      #print newname
      if not os.path.exists(newname):
        #print "doesnt exist"
        tn=getThumbnail(f, (n[0],n[0]) )
        tn.save(newname)
  try:
    while True:
      jpgFiles.remove(None)
  except ValueError:
    pass
      
  files = zip(jpgFiles, photoDates, isMovie)
  files.sort(lambda a, b: cmp(a[1], b[1]))
  return files

def convertGpsToReal(position, sign):
  deg = position.split(".")[0][:-2]
  min = float(position[len(deg):])
  if sign in "SW":
    deg = "-"+deg
  #print "dm",deg,min
  return int(deg)+min/60.0

def getGpsData():
  good = 0
  total = 0
  bad = 0
  maxrow = 0
  currow = 0
  lastpoint = None
  gpsFiles = filter((lambda x: x.lower().endswith(".nmea")), os.listdir("."))
  points=[]
  for filename in gpsFiles:
    lines=open(filename).readlines()
    starttag=lines[0].split(",")[0]
    endtag=None
    for l in lines[1:]:
      if l.startswith(starttag):
        break
      endtag=l.split(",")[0]

    gpspoint={}
    for l in lines:
      tags = l.split(",")
      if tags[0] == "$GPGGA":
        gpspoint["time"] = tags[1].split(".")[0]
        gpspoint["lat"] = convertGpsToReal(tags[2], tags[3])
        gpspoint["lon"] = convertGpsToReal(tags[4], tags[5])

        gpspoint["altitude"] = int(round(float(tags[9])))
        if not 0 < gpspoint["altitude"] < 20000:
          #print "funny altitude:", gpspoint["altitude"]
          gpspoint["altitude"] = None
        
      elif tags[0] == "$GPRMC":
        gpspoint["date"] = tags[9]
      elif tags[0] == "$GPGSA":
        gpspoint["pdop"] = tags[15]

      if l.startswith(endtag):
        #print "datime", gpspoint["date"]+gpspoint["time"]
        t=mktime(strptime(gpspoint["date"]+gpspoint["time"], "%d%m%y%H%M%S"))
        gpspoint["secTime"] = t+timeDifference
        if startTime < gpspoint["secTime"] < endTime:
          total += 1
          #check accuracy
          if (float(gpspoint["pdop"]) < 10):
            points.append(gpspoint)
            gpspoint['break'] = False
            if lastpoint and lastpoint['secTime'] + 60 < gpspoint['secTime']:
              #print 'td',  (-lastpoint['secTime'] + gpspoint['secTime'])
              lastpoint['break'] = True
            lastpoint = gpspoint
            good += 1
            maxrow = max(maxrow, currow)
            currow = 0;
            #print "ALTITUDE:", gpspoint["altitude"]
          else:
            #print "too high pdop", gpspoint["pdop"]
            bad += 1
            currow += 1
        gpspoint={}



  #sort points by time
  #a = str(points)
  points.sort(lambda a, b: cmp(a["secTime"], b["secTime"]))
  #print "CMP:",cmp(str(points), a)
  print "GPS points: total, good, bad, maxrow", total, good, bad, maxrow
  return points


def getLocation(time, points):
  tag = "secTime"
  next = None
  previous = None

  previous = points[-2]
  next = points[-1]
  timeDiff = 1000
  for i in range(1,len(points)):
    if time < points[i]["secTime"]:
      previous = points[i-1]
      next = points[i]
      timeDiff = min(abs(time-next["secTime"]), abs(time-previous["secTime"]))
      if timeDiff > 30:
        print "Large TimeDifference (unaccurate location) at ",time, " :", timeDiff
      break
  #print "distance", distance(previous, next)
  tbp = abs(next[tag] - previous[tag]) #time between points
  #print tbp
  #print (next[tag]-time), next["lat"], (time-previous[tag]), previous["lat"], tbp, (next[tag] - previous[tag])
  lon = ((time-previous[tag])*next["lon"] + (next[tag]-time)*previous["lon"])/tbp
  lat = ((time-previous[tag])*next["lat"] + (next[tag]-time)*previous["lat"])/tbp
  alt = "unknown"
  if (next["altitude"] and previous["altitude"]):
    alt = int(((next[tag]-time)*next["altitude"] + (time-previous[tag])*previous["altitude"])/tbp)
  elif next["altitude"]:
    alt = next["altitude"]
  elif previous["altitude"]:
    alt = previous["altitude"]
  
  #print tbp, next[tag]-time, time-previous[tag]
  #print "prev:",previous["lon"]
  #print lon, lat
  #print next["lon"]
  
  return {"lat": lat, "lon": lon, "altitude": alt, "timeDiff": timeDiff}

def timeToHtml(time):
  month = str(int(strftime("%m", localtime(time))))
  day = str(int(strftime("%d", localtime(time))))
  return strftime("%a,&nbsp;"+day+"."+month+".%Y&nbsp;%H:%M:%S", localtime(time))

def getGroupLoc(group):
  lat= lon= alt= 0.0
  for m in group:
    lat += m[1]["lat"]
    lon += m[1]["lon"]
    if m[1]["altitude"] == "unknown":
      alt = "unknown"
    if alt != "unknown":
      alt += m[1]["altitude"]

  lat /= len(group)
  lon /= len(group)
  if alt != "unknown":
    alt /= len(group)
  result = { "lat": lat, "lon": lon, "altitude": alt }
  group[0][1]["grouplocation"] = result
  group[0][1]["groupsize"] = len(group)
  return result

def getMinDist(groups):
  if len(groups) < 2:
    return 100000, None, None
  minDist = 100000
  minA, minB = None, None
  for a in groups:
    for b in groups:
      if (a != b and minDist > distance(a,b)):
        minDist = distance(a, b)
        minA, minB = a, b

  return minDist, minA, minB


tnsize = 152
def getTnSize(filename):
  image = Image.open("webpages/"+directory+filename)
  width, height = image.size[0], image.size[1]
  if (width > height):
    height = height * tnsize/width
    width = tnsize
  else:
    width = width * tnsize/height
    height = tnsize
  #print width, height
  return str(width), str(height)

def addPhotos(filenames, gpsdata):
  scrollBarHtml = ""
  photoHtml     = ""
  a = 0

  locations = []

  for f in filenames:
    gpspoint = getLocation(f[1], gpsdata)
    point = gpspoint["lat"],gpspoint["lon"]
    locations.append( [ (f[0],gpspoint, f[1], f[2]) ] )
    image = Image.open("webpages/"+directory+f[0])
    width, height = getTnSize(f[0])
    
    photoHtml += "createPhotoInfoWin(new GLatLng" + str(point) + ', "'
    photoHtml += str(f[0]) +'", "'+ timeToHtml(f[1])+'", "'+ str(gpspoint["altitude"])
    if type(gpspoint["altitude"]) == type(1):
      photoHtml += "m"
    photoHtml += '", ' + width + ", " + height + ' );\n'
    
    if (f[2]): # this is a movie
      scrollBarHtml += '<p style="border: 2px solid">'
    if gpspoint["timeDiff"] > 30:
      scrollBarHtml += "Location possibly UNACCURATE ("+str(gpspoint["timeDiff"])+")<br>"
    scrollBarHtml += "Time: "+ timeToHtml(f[1])+" <br>Altitude:"+str(gpspoint["altitude"])
    if type(gpspoint["altitude"]) == type(1):
      scrollBarHtml += "m"
    scrollBarHtml += "<br>"
    if (f[2]): # this is a movie
      param = '="' + surl + 'flvplayer.swf?file='+ target + f[0].split(".")[0]+'.flv&autoStart=false"'
      scrollBarHtml += 'THIS IS A MOVIE<br>'
      scrollBarHtml += '<a href="javascript:void(0)" onclick="openInfoWin(' +str(a)
      scrollBarHtml += ', true)";>See location</a><br>\n'
      scrollBarHtml += '<object style="margin: 5px"'
      scrollBarHtml += ' type="application/x-shockwave-flash" width="296" height="242" wmode="transparent"'
      scrollBarHtml += ' data' + param + '>\n<param name="movie" value' + param + ' />'
      scrollBarHtml += '\n<param name="wmode" value="transparent" /></object>\n'
      scrollBarHtml += '</p>'
    else:
      scrollBarHtml += '<a href=# onclick=openWin("' +f[0] + '") >Open bigger photo (in new window)</a><br>'
      scrollBarHtml += '<a href="javascript:void(0)" onclick="openInfoWin(' +str(a)
      scrollBarHtml += ', true)";><img src="small_' + f[0] + '"></a><br><br>\n'
    
    a = a+1

  print "Creating groups...",
  while True:
    dist, a, b = getMinDist(locations)  
    #print "LOCATIONS:", len(locations), "Asize", len(a), "Bsize", len(b), "mindist", dist
    print ".",
    if (dist > 0.1):
      print "\nGroups formed, continuing..."
      break

    locations.remove(a)
    locations.remove(b)
    locations.append(a+b)

  largest = 0
  for l in locations:
    largest = max(largest, len(l))
  
  print "Groups:",len(locations), "Largest group:",largest

  #print locations[0]

  for l in locations:
    gpspoint = getGroupLoc(l)
    
    point = "new GLatLng"+str((gpspoint["lat"],gpspoint["lon"]))
    html = '[ "'

    if (len(l) > 1):
      html += '<table border=1 ><tr>'
    for i in range(len(l)):
      img = l[i]
      if i > 0 and i % 4 == 0:
        html += '</tr></table>", "<table border=1 ><tr>'
      if i % 4 == 2:
        html += "</tr><tr>"
      if (len(l) > 1):
        html += "<td>"
      width, height = getTnSize(img[0])
      html += '<img width='+ width +' height='+ height +' src=\\"small_' + img[0] + '\\"><br>'
      html += timeToHtml(img[2]) +"<br>Altitude: "+str(img[1]["altitude"])+ 'm<br>'
      if img[3]: # this is a movie
        html += 'Movie'
      else:
        html += '<a href=# onclick=openWin(\\"' +img[0] + '\\") >Open in new window</a>'
      if (len(l) > 1):
        html += "</td>"


    if (len(l) > 1):
      html += '</tr></table>'
    html += '" ]'
    icon = surl+'marker'+str(len(l))+'.png'
    photoHtml += "map.addOverlay(createMarker("+ point +','+ html +', "'+ icon +'"));\n'
    
    files = map(lambda x: x[0],l)
    alts = map(lambda x: x[1]["altitude"],l)
    points = map(lambda x: [x[1]["lat"], x[1]["lon"]],l)

  scrollBarHtml = scrollBarHtml[:-9]

  # add startingpoint and endingpoint markers
  for p in [ ("Start",gpsdata[0]), ("End", gpsdata[-1]) ]:
    infohtml = p[0]+" of route<br>";
    infohtml += "Time: "+ timeToHtml(p[1]["secTime"])+" <br>Altitude:"+str(p[1]["altitude"])+"m"
    photoHtml += "map.addOverlay(createMarker(new GLatLng" + str((p[1]["lat"],p[1]["lon"])) + \
                 ', ["'+infohtml+'"], "' + surl + p[0] + '.png" ));\n';

  return photoHtml, scrollBarHtml

def dotLineIntersection(x0,y0, x1,y1, x,y):   # p0, p1, p):
    if not ( x0 <= x <= x1 or x1 <= x <= x0):
      return None
    if not ( y0 <= y <= y1 or y1 <= y <= y0):
      return None

    if(x1 - x0) == 0:
        return {"lat":x0, "lon":y}
    elif(y1 - y0) == 0:
        return {"lat":x, "lon":y0}
    tg = -1 / ((y1 - y0) / (x1 - x0));
    rx = (x1 * (x * tg - y + y0) + x0 * (x * - tg + y - y1)) / (tg * (x1 - x0) + y0 - y1)
    ry = tg * rx - tg * x + y
    return {"lat":rx, "lon":ry}

def usefulness(p0, p1, p):
  usefullness = distance(p0,p)
  lp = dotLineIntersection(p0["lat"], p0["lon"], p1["lat"], p1["lon"], p["lat"], p["lon"])
  if lp:
    usefullness = min(usefullness, distance(lp, p)*3)
  else:
    lp = p
  #print "dist, lp, usefullness", distance(p0,p), distance(lp, p), usefullness
  return usefullness


def getUsefulPoints(points):
  points = points[:]
  points[0] ['usefulness'] = 1000
  points[-1]['usefulness'] = 1000

  for i in range(1, len(points)-1):
    points[i]['usefulness'] = usefulness(points[i-1], points[i+1], points[i])

  print "Removing useless GPS points..."
  while len(points) > 200:
    #remove least useful point
    #remember to update usefulness values of nearby points
    worst_u = 1000
    for p in points:
      if p['usefulness'] < worst_u:
        worstpoint = p
        worst_u = p['usefulness']
    index = points.index(worstpoint)
    #print "removing", index, worstpoint['usefulness']
    points.remove(worstpoint)
    for i in [index-1, index]:
      if 0 < i < len(points)-1:
        old = points[i]['usefulness']
        points[i]['usefulness'] = max(old, usefulness(points[i-1], points[i+1], points[i]))
        #print "old, new, diff", old, points[i]['usefulness'], (points[i]['usefulness']-old)
      #else:
        #print "i", i

  return points      

def breakBetween(a, b, points):
  i = 0
  
  while(True):
    if points[i] == a:
      break
    i += 1
  while(True):
    if points[i] == b:
      return None
    if points[i]['break']:
      #print "\t",a['secTime'], b['secTime'], points[i]['secTime']
      return i
    i += 1



startTime = 0
endTime = 2**32
directory = ''

if len(argv) > 3:
  startTime = mktime(strptime(str(argv[1]), "%d.%m.%Y_%H:%M:%S"))
  endTime   = mktime(strptime(str(argv[2]), "%d.%m.%Y_%H:%M:%S"))
  directory = argv[3]+"/"
  #print startTime, endTime, directory


def mkdir(dir):
  try:
    os.mkdir(dir)
    print "created directory", dir
  except:
    pass



mkdir("webpages")
mkdir("webpages/"+directory)

if (delete):
  os.system("rm webpages/*")
  os.system("rm webpages/" + directory + "/*")
 
pathHtml = ""
googleMapsKey += '"'

print "loading GPS files..."
points = getGpsData()

for p in points:
  if type(p) == type(()):
    print p

print "getting useful points..."
usefulPoints = getUsefulPoints(points)

print "Total points:", len(points), "useful points:",len(usefulPoints)

total = 0;
for i in range(len(usefulPoints)-1):
  fr = usefulPoints[i]
  to = usefulPoints[i+1]
  d = distance(fr, to)
  total += d
print "total length of track:", total, "km"

for i in range(len(usefulPoints)-1):
  p = usefulPoints[i]
  nextp = usefulPoints[i+1]
  pathHtml += "new GPoint(" + str(p["lon"]) +", "+ str(p["lat"]) + "),\n"
  bp = breakBetween(p, nextp, points)
  if bp and distance(p,nextp) > 0.2:
    print "break in gps data, time:",(nextp['secTime']-p['secTime']), 's, distance:', distance(p,nextp),"km"   
    #pathHtml += "new GPoint(" + str(points[bp]["lon"]) +", "+ str(points[bp]["lat"]) + "),\n"
    pathHtml = pathHtml[:-2]
    pathHtml += '   ],  "#ff0000", 3, 1.0));\n'
    pathHtml += '    map.addOverlay( new GPolyline([\n'
    start = None
    for n in range(len(points)):
      if points[n] == nextp:
        if not start:
          start = n
        break;
      if points[n]['break']:
        start = n+1
    #print "p, bp, start, nextp", bp, start, p['secTime'], points[bp]['secTime'], points[start]['secTime'], nextp['secTime']
    #pathHtml += "new GPoint(" + str(points[start]["lon"]) +", "+ str(points[start]["lat"]) + "),\n"
    

p=usefulPoints[-1]
pathHtml += "new GPoint(" + str(p["lon"]) +", "+ str(p["lat"]) + "),\n"
pathHtml = pathHtml[:-2]

jpgFiles = generateImages()
photosHtml, scrollBarHtml = addPhotos(jpgFiles, points)

minLat = minLon = 1000
maxLat = maxLon = -1000
for p in points:
  minLat = min(minLat, p["lat"])
  minLon = min(minLon, p["lon"])
  maxLat = max(maxLat, p["lat"])
  maxLon = max(maxLon, p["lon"])
  
print "minmaxes:",minLat, minLon, maxLat, maxLon

centerLon = (minLon+maxLon)/2
centerLat = (minLat+maxLat)/2

latDist = distance({"lat": minLat, "lon": centerLon}, {"lat": maxLat, "lon": centerLon})
lonDist = distance({"lat": centerLat, "lon": minLon}, {"lat": centerLat, "lon": maxLon})

print "latDist, lonDist", latDist, lonDist

print len(jpgFiles), "pictures,", len(usefulPoints), "of", len(points), " gps points used"

input=open("index.htpy").read()

indexhtml = eval('"""'+input+'"""')

out=open("webpages/"+directory+"index.html",'w')
out.write(indexhtml)

# That's all, folks
